@@ -8,13 +8,13 @@ import multiprocessing
 
 from . import utils
 from . import stats
-from .utils import MapResults
+from . import pslogger
 
 pool = None
 our_ncores = None
 
 
-def init(ncores=None, verbose=None):
+def init(system_kwargs, ncores=None, **kwargs):
     global pool
     global our_ncores
     if pool:  # yes we can be called multiple times  # pragma: no cover
@@ -22,7 +22,7 @@ def init(ncores=None, verbose=None):
 
     if ncores is None:
         ncores = multiprocessing.cpu_count()
-    if verbose:
+    if system_kwargs['verbose']:
         print('initializing multiprocessing pool with {} processes'.format(ncores), file=sys.stderr)
     pool = multiprocessing.Pool(processes=ncores)
     our_ncores = ncores
@@ -80,6 +80,7 @@ def do_work_wrapper(func, system_kwargs, user_kwargs, psets):
                 print('saw an exception in the worker function', file=sys.stderr)
                 print('it was working on', json.dumps(pset, sort_keys=True), file=sys.stderr)
                 traceback.print_exc()
+                user_ret['traceback'] = traceback.format_exc()
             ret.append([user_ret, system_ret])
         return ret
     except Exception as e:
@@ -90,6 +91,7 @@ def do_work_wrapper(func, system_kwargs, user_kwargs, psets):
         # cannot increment progress[failures] here because we are in the child & it is not returned
         # fake up a single return value
         user_ret = {'pset': psets[0], 'exception': repr(e)}
+        user_ret['traceback'] = traceback.format_exc()
         return [[user_ret, {}]]
 
 
@@ -98,32 +100,36 @@ def callback(out_func, system_stats, system_kwargs, user_kwargs, ret):
     utils.handle_return_common(out_func, ret, system_stats, system_kwargs, user_kwargs)
 
 
-def error_callback(out_func, system_stats, system_kwargs, user_kwargs, ret):
-    system_kwargs['outstanding'] -= 1
-    system_kwargs['progress'].failures += 1
-    print('error_callback, exception is', repr(ret), file=sys.stderr)
+def error_callback(e):
+    # not sure I have ever seen this called
+    print('error_callback, exception is', repr(e), file=sys.stderr)
+    raise ValueError('error_callback, exception is '+repr(e))
 
 
 def progress_until_fewer(cores, factor, out_func, system_stats, system_kwargs, user_kwargs, group_size):
     verbose = system_kwargs['verbose']
 
+    count = 0
     while system_kwargs['outstanding'] > cores*factor:
         time.sleep(0.1)
-        if verbose > 1:
-            system_stats.bingo()
+        count += 1
+        if count % 10 == 0 and verbose > 2:
+            print('looping in the progress loop', file=sys.stderr)
 
     return group_size
 
 
-def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None, out_subdirs=None,
-        progress_dt=60., group_size=None, name='default', verbose=None, **kwargs):
+def map(func, psets, out_func=None, system_kwargs=None, user_kwargs=None, chdir=None, outfile=None, out_subdirs=None,
+        progress_dt=None, group_size=None, name='default', **kwargs):
+
+    verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
 
     if utils.psets_empty(psets):
         return
 
-    verbose = verbose or 0
-
-    psets, system_stats, system_kwargs = utils.map_prep(psets, name, chdir, outfile, out_subdirs, verbose, **kwargs)
+    psets, system_stats, system_kwargs = utils.map_prep(psets, name, system_kwargs, chdir, outfile,
+                                                        out_subdirs, progress_dt=progress_dt, **kwargs)
 
     progress = system_kwargs['progress']
     cores = current_core_count()
@@ -150,23 +156,20 @@ def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None,
 
     while True:
         while system_kwargs['outstanding'] <= cores * factor:
-            with stats.record_wallclock('get_pset_group', obj=system_stats):
-                pset_group, pset_index = utils.get_pset_group(psets, pset_index, group_size)
+            pset_group, pset_index = utils.get_pset_group(psets, pset_index, group_size)
             if len(pset_group) == 0:
                 break
 
             pset_group, pset_ids = utils.make_pset_ids(pset_group)
             system_kwargs['pset_ids'].update(pset_ids)
 
-            with stats.record_wallclock('multiprocessing.apply_async', obj=system_stats):
-                pool.apply_async(do_work_wrapper,
-                                 (func, worker_system_kwargs, user_kwargs, pset_group),
-                                 {}, callback_partial, error_callback_partial)
-            if verbose > 1:
-                system_stats.bingo()
+            pool.apply_async(do_work_wrapper,
+                             (func, worker_system_kwargs, user_kwargs, pset_group),
+                             {}, callback_partial, error_callback_partial)
             system_kwargs['outstanding'] += 1
             progress.started += len(pset_group)
-            utils.report_progress(system_kwargs)
+            progress.report(verbose)
+            system_stats.report(vstats, other_fd=pslogger.logfd)
 
         if pset_index >= len(psets):
             break
@@ -180,19 +183,4 @@ def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None,
 
     progress_until_fewer(cores, 0, out_func, system_stats, system_kwargs, user_kwargs, group_size)
 
-    if verbose:
-        print('finished getting results', file=sys.stderr)
-        sys.stderr.flush()
-
-    utils.finalize_progress(system_kwargs)
-    utils.report_progress(system_kwargs, final=True)
-
-    system_stats.print_percentiles(name)
-    missing = list(system_kwargs['pset_ids'].values())
-
-    if 'results' in system_kwargs:
-        results = system_kwargs['results'].finalize()
-    else:
-        results = None
-
-    return MapResults(results, missing, system_kwargs['progress'], system_stats)
+    return utils.map_finalize(name, system_kwargs, system_stats)
