@@ -5,11 +5,11 @@ import json
 import time
 
 import ray
-import pyarrow
+#import pyarrow
 
 from . import utils
 from . import stats
-from .utils import MapResults
+from . import pslogger
 
 
 def read_ray_config():
@@ -21,25 +21,24 @@ def read_ray_config():
     return address, password
 
 
-def init(verbose=False, **kwargs):
+def init(system_kwargs, ncores=None, **kwargs):
     ray_kwargs = {}
 
-    if 'ncores' in kwargs:
-        # what does num_cpus actually do if the the cluster pre-exists?
-        ray_kwargs['num_cpus'] = kwargs['ncores']
-        kwargs.pop('ncores')
+    if ncores:
+        ray_kwargs['num_cpus'] = ncores
 
+    # should allow these to be kwargs
     address, password = read_ray_config()
     kwargs['address'] = address
     kwargs['redis_password'] = password
 
     if 'ignore_reinit_error' not in kwargs:
-        kwargs['ignore_reinit_error'] = True  # XXX needed for our test infra
+        kwargs['ignore_reinit_error'] = True  # needed for testing
 
     if os.environ.get('RAY_LOCAL_MODE', False):
         kwargs['local_mode'] = True
 
-    # XXX if the cluster does not pre-exist, should we create it?
+    # should we create a ray head if there is no cluster?
     ray.init(**kwargs)
 
 
@@ -87,11 +86,13 @@ def do_work_wrapper(func, system_kwargs, user_kwargs, psets):
             print('saw an exception in the worker function', file=sys.stderr)
             print('it was working on', json.dumps(pset, sort_keys=True), file=sys.stderr)
             traceback.print_exc()
+            user_ret['traceback'] = traceback.format_exc()
         ret.append([user_ret, system_ret])
     return ret
 
 
 def handle_return(out_func, ret, system_stats, system_kwargs, user_kwargs):
+    verbose = system_kwargs['verbose']
     progress = system_kwargs['progress']
 
     try:
@@ -99,17 +100,19 @@ def handle_return(out_func, ret, system_stats, system_kwargs, user_kwargs):
             ret = ray.get(ret)
     except Exception as e:
         # RayTaskError has been seen here
-        print('\nSurprised by exception {} getting a result,\n'
-              'an unknown number of results lost\n'.format(e), file=sys.stderr)
+        err = '\nSurprised by exception {} ray.getting a result, an unknown number of results lost\n'.format(e)
+        print(err, file=sys.stderr)
         traceback.print_exc()
         sys.stderr.flush()
-        progress.failures += 1
-        utils.report_progress(system_kwargs)
+        pslogger.log(err)
+        pslogger.log(traceback.format_exc())
+        progress.report(verbose)
         return
 
     utils.handle_return_common(out_func, ret, system_stats, system_kwargs, user_kwargs)
 
 
+'''
 def check_serialized_size(args, factor=1.2):
     big_data = 10 * 1024 * 1024 * 1024  # TODO: make this dynamic with cluster resources
     cores = current_core_count()
@@ -124,23 +127,27 @@ def check_serialized_size(args, factor=1.2):
         print('due to small in-flight data size, goosing factor by 2x', file=sys.stderr)
         factor *= 2
     return factor
+'''
 
 
 def progress_until_fewer(futures, cores, factor, out_func, system_stats, system_kwargs, user_kwargs, group_size):
     verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
 
     while len(futures) > cores*factor:
-        with stats.record_wallclock('ray.wait', obj=system_stats):
-            t0 = time.time()
-            done, pending = ray.wait(futures, num_returns=len(futures), timeout=1)
-            elapsed = time.time() - t0
+        t0 = time.time()
+        done, pending = ray.wait(futures, num_returns=len(futures), timeout=1)
+        elapsed = time.time() - t0
 
         print_nums = False
         if elapsed < 0.8:  # pragma: no cover
-            print('something bad happened in ray.wait, normally it takes 2.0 seconds, but it took', elapsed, file=sys.stderr)
-            print_nums = True
+            if len(pending):
+                # only observed at the end, when pending == 0
+                print('something bad happened in ray.wait, normally it takes 2.0 seconds, but it took', elapsed, file=sys.stderr)
+                print_nums = True
 
         if len(futures) != len(done) + len(pending):  # pragma: no cover
+            # never observed
             print('something bad happened in ray.wait, counts do not add up:')
             print_nums = True
 
@@ -159,29 +166,29 @@ def progress_until_fewer(futures, cores, factor, out_func, system_stats, system_
             for ret in done:
                 handle_return(out_func, ret, system_stats, system_kwargs, user_kwargs)
 
-        if verbose > 1:
-            system_stats.bingo()
-
         new_cores = current_core_count()
-        if new_cores != cores:
-            print('core count changed from {} to {}'.format(cores, new_cores), file=sys.stderr)
+        if new_cores != cores:  # not tested
+            if verbose:
+                print('core count changed from {} to {}'.format(cores, new_cores), file=sys.stderr)
+                sys.stderr.flush()
             cores = new_cores
-            sys.stderr.flush()
 
         # dynamic group_size adjustment
 
     return futures, cores, group_size
 
 
-def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None, out_subdirs=None,
-        progress_dt=60., group_size=None, name='default', verbose=None, **kwargs):
+def map(func, psets, out_func=None, system_kwargs=None, user_kwargs=None, chdir=None, outfile=None, out_subdirs=None,
+        progress_dt=None, group_size=None, name='default', **kwargs):
+
+    verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
 
     if utils.psets_empty(psets):
         return
 
-    verbose = verbose or 0
-
-    psets, system_stats, system_kwargs = utils.map_prep(psets, name, chdir, outfile, out_subdirs, verbose, **kwargs)
+    psets, system_stats, system_kwargs = utils.map_prep(psets, name, system_kwargs, chdir, outfile,
+                                                        out_subdirs, progress_dt=progress_dt, **kwargs)
     if 'chdir' not in system_kwargs:
         # ray workers default to ~
         system_kwargs['chdir'] = os.getcwd()
@@ -218,20 +225,17 @@ def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None,
 
     while True:
         while len(futures) <= cores * factor:
-            with stats.record_wallclock('get_pset_group', obj=system_stats):
-                pset_group, pset_index = utils.get_pset_group(psets, pset_index, group_size)
+            pset_group, pset_index = utils.get_pset_group(psets, pset_index, group_size)
             if len(pset_group) == 0:
                 break
 
             pset_group, pset_ids = utils.make_pset_ids(pset_group)
             system_kwargs['pset_ids'].update(pset_ids)
 
-            with stats.record_wallclock('ray.remote', obj=system_stats):
-                futures.append(do_work_wrapper.remote(func, worker_system_kwargs, user_kwargs, pset_group))
-            if verbose > 1:
-                system_stats.bingo()
+            futures.append(do_work_wrapper.remote(func, worker_system_kwargs, user_kwargs, pset_group))
             progress.started += len(pset_group)
-            utils.report_progress(system_kwargs)
+            progress.report(verbose)
+            system_stats.report(vstats, other_fd=pslogger.logfd)
 
         if pset_index >= len(psets):
             break
@@ -245,19 +249,4 @@ def map(func, psets, out_func=None, user_kwargs=None, chdir=None, outfile=None,
 
     progress_until_fewer(futures, cores, 0, out_func, system_stats, system_kwargs, user_kwargs, group_size)
 
-    if verbose:
-        print('finished getting results', file=sys.stderr)
-        sys.stderr.flush()
-
-    utils.finalize_progress(system_kwargs)
-    utils.report_progress(system_kwargs, final=True)
-
-    system_stats.print_percentiles(name)
-    missing = list(system_kwargs['pset_ids'].values())
-
-    if 'results' in system_kwargs:
-        results = system_kwargs['results'].finalize()
-    else:
-        results = None
-
-    return MapResults(results, missing, system_kwargs['progress'], system_stats)
+    return utils.map_finalize(name, system_kwargs, system_stats)
