@@ -5,17 +5,28 @@ import math
 import random
 import keyword
 import resource
+import os
 
 import pandas as pd
 from pandas_appender import DF_Appender
 
 from . import stats
+from . import pslogger
 
 
 class MapProgress(object):
     '''Class to track progress of a map()'''
     # would be perfect as a dataclass, once python 3.7 is our minimum
-    def __init__(self, d={}):
+    def __init__(self, name, d={}, verbose=1, progress_dt=None):
+        self.name = name
+        self.print_last = time.time()
+        if progress_dt is not None:
+            self.print_dt = progress_dt
+        else:
+            self.print_dt = self.pick_dt(verbose)
+        self.print_log_last = time.time()
+        self.print_log_dt = 30  # hardwired
+
         self.total = d.get('total', 0)
         self.started = d.get('started', 0)
         self.finished = d.get('finished', 0)
@@ -23,7 +34,43 @@ class MapProgress(object):
         self.exceptions = d.get('exceptions', 0)
 
     def __str__(self):
-        return ', '.join([k+': '+str(v) for k, v in vars(self).items()])
+        attrs = ('total', 'started', 'finished', 'failures', 'exceptions')
+        return ', '.join([k+': '+str(getattr(self, k)) for k in attrs])
+
+    def pick_dt(self, verbose):
+        if verbose >= 3:
+            return 0
+        elif verbose >= 2:
+            return 1
+        elif verbose == 1:
+            return 30
+        else:
+            return 1000000
+
+    def finalize(self, verbose, missing):
+        failures = self.failures
+        actual_failures = len(missing)
+
+        # fixup wrapper failures
+        if actual_failures > failures:
+            print('correcting failure count from {} to {}'.format(failures, actual_failures), file=sys.stderr)
+            self.failures = actual_failures
+        elif actual_failures < failures:
+            print('can\'t happen! missing pset_ids {} less than failures {}'.format(actual_failures, failures), file=sys.stderr)
+        else:
+            if verbose > 1 and failures > 0:
+                print('failures equal to actual failures, hurrah', file=sys.stderr)
+
+    def report(self, verbose=None, final=None, other_fd=None):
+        t = time.time()
+        if final or t - self.print_last > self.print_dt:
+            self.print_last = t
+            print(self.name, 'progress:', str(self), file=sys.stderr)
+            sys.stderr.flush()
+        if other_fd and (final or t - self.print_log_last > self.print_log_dt):
+            self.print_log_last = t
+            print(self.name, 'progress:', str(self), file=other_fd)
+            other_fd.flush()
 
 
 class MapResults(object):
@@ -103,23 +150,17 @@ class PDFWrapper(object):
 '''
 
 
-def report_progress(system_kwargs, final=False):
-    t = time.time()
-    verbose = system_kwargs['verbose']
-
-    force = bool(final or verbose > 1)
-
-    if force or t - system_kwargs['progress_last'] > system_kwargs['progress_dt']:
-        system_kwargs['progress_last'] = t
-        print(system_kwargs['name'], 'progress:', str(system_kwargs['progress']),
-              file=sys.stderr)
-
-        if final and verbose and system_kwargs['pset_ids']:
+def report_missing(missing, verbose, other_fd=None):
+    if missing:
+        if verbose > 1:
             print('missing psets:', file=sys.stderr)
-            for pset_id, pset in system_kwargs['pset_ids'].items():
+        if other_fd:
+            print('missing psets:', file=other_fd)
+        for pset in missing:
+            if verbose > 1:
                 print(' ', pset, file=sys.stderr)
-
-        sys.stderr.flush()
+            if other_fd:
+                print(' ', pset, file=other_fd)
 
 
 def remaining(system_kwargs):
@@ -151,16 +192,24 @@ def psets_prep(psets):
     return psets
 
 
-def map_prep(psets, name, chdir, outfile, out_subdirs, verbose, keep_results=True, **kwargs):
-    print('starting work on', name, file=sys.stderr)
-    sys.stderr.flush()
+def map_prep(psets, name, system_kwargs, chdir, outfile, out_subdirs, keep_results=True, progress_dt=None, **kwargs):
+    verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
+
+    if verbose > 0:
+        print('paramsurvey.map starting work on', name, file=sys.stderr)
+        sys.stderr.flush()
+    pslogger.log('paramsurvey.map starting work on '+name)
+    pslogger.log('paramsurvey.map system_kwargs '+repr(system_kwargs))
 
     psets = psets_prep(psets)
+    if system_kwargs['limit'] >= 0:
+        psets = psets.iloc[:system_kwargs['limit']]
+
+    system_kwargs['progress'] = MapProgress(name, {'total': len(psets)}, verbose=verbose, progress_dt=progress_dt)
 
-    system_kwargs = {'progress': MapProgress({'total': len(psets)})}
     if keep_results:
         system_kwargs['results'] = DF_Appender(ignore_index=True)
-
     if chdir:
         system_kwargs['chdir'] = chdir
     if outfile:
@@ -169,20 +218,44 @@ def map_prep(psets, name, chdir, outfile, out_subdirs, verbose, keep_results=Tru
         system_kwargs['out_subdirs'] = out_subdirs
     if name:
         system_kwargs['name'] = name
-    system_kwargs['verbose'] = verbose or 0
-
     if 'raise_in_wrapper' in kwargs:
         system_kwargs['raise_in_wrapper'] = kwargs['raise_in_wrapper']
 
     system_kwargs['pset_ids'] = {}
 
-    system_stats = stats.PerfStats()
-    system_kwargs['progress_last'] = 0.
-    system_kwargs['progress_dt'] = 0.
+    system_stats = stats.PerfStats(vstats=vstats)
+
+    pslogger.log('paramsurvey.map pset count {}, pset columns {}'.format(len(psets), list(psets.columns.values)))
 
     return psets, system_stats, system_kwargs
 
 
+def map_finalize(name, system_kwargs, system_stats):
+    verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
+
+    if verbose:
+        print('finished getting results', file=sys.stderr)
+        sys.stderr.flush()
+
+    progress = system_kwargs['progress']
+    progress.finalize(verbose, system_kwargs['pset_ids'])
+    progress.report(final=True, other_fd=pslogger.logfd)
+    system_stats.report(vstats, final=True, other_fd=pslogger.logfd)
+
+    missing = list(system_kwargs['pset_ids'].values())
+    report_missing(missing, verbose, other_fd=pslogger.logfd)
+
+    if 'results' in system_kwargs:
+        results = system_kwargs['results'].finalize()
+    else:
+        results = None
+
+    pslogger.log('paramsurvey.map returning results')
+
+    return MapResults(results, missing, system_kwargs['progress'], system_stats)
+
+
 def flatten_results(results):
     seen_pset_keys = set()
     seen_result_keys = set()
@@ -218,25 +291,11 @@ def make_pset_ids(psets):
     return ret, pset_ids
 
 
-def finalize_progress(system_kwargs):
-    progress = system_kwargs['progress']
-    failures = progress.failures
-    actual_failures = len(system_kwargs['pset_ids'])
-
-    # needed to fixup wrapper failures
-    if actual_failures > failures:
-        print('correcting failure count from {} to {}'.format(failures, actual_failures), file=sys.stderr)
-        progress.failures = actual_failures
-    elif actual_failures < failures:
-        print('can\'t happen! missing pset_ids {} less than failures {}'.format(actual_failures, failures), file=sys.stderr)
-    else:
-        if system_kwargs['verbose'] > 1 and failures > 0:
-            print('failures equal to actual failures, hurrah', file=sys.stderr)
-
-
 def handle_return_common(out_func, ret, system_stats, system_kwargs, user_kwargs):
     progress = system_kwargs['progress']
     verbose = system_kwargs['verbose']
+    vstats = system_kwargs['vstats']
+
     for user_ret, system_ret in ret:
         if 'result' in user_ret and not isinstance(user_ret['result'], dict) and user_ret['result'] is not None:
             # fake an exception, make this case look like other failures
@@ -253,9 +312,17 @@ def handle_return_common(out_func, ret, system_stats, system_kwargs, user_kwargs
         if 'exception' in user_ret:
             progress.failures += 1
             progress.exceptions += 1
+            pslog = 'saw exception in worker: ' + user_ret['exception']
+            if verbose > 0:
+                print(pslog, file=sys.stderr)
+            pslog += '\n  pset: '+repr(system_kwargs['pset_ids'][pset_id])
             system_kwargs['pset_ids'][pset_id]['_exception'] = user_ret['exception']
-            if verbose > 1:
-                print('saw exception', user_ret['exception'], file=sys.stderr)
+            if 'traceback' in user_ret:
+                system_kwargs['pset_ids'][pset_id]['_traceback'] = user_ret['traceback']
+                pslog += '\n  traceback:\n' + user_ret['traceback']
+                if verbose > 1:
+                    print('saw traceback', user_ret['traceback'], file=sys.stderr)
+            pslogger.log(pslog)
         else:
             del system_kwargs['pset_ids'][pset_id]
             user_ret['pset'].pop('_pset_id', None)
@@ -273,7 +340,47 @@ def handle_return_common(out_func, ret, system_stats, system_kwargs, user_kwargs
         if out_func:
             out_func(user_ret, system_kwargs, user_kwargs)
 
-    report_progress(system_kwargs)
+    progress.report(verbose)
+    system_stats.report(vstats, other_fd=pslogger.logfd)
+
+
+def initialize_kwargs(global_kwargs, kwargs):
+    for k, v in global_kwargs.items():
+        value = None
+        if v['env'] in os.environ:
+            value = os.environ[v['env']]
+            v['strong'] = True
+        elif kwargs.get(k) is not None:
+            value = kwargs.get(k)
+        else:
+            value = v.get('default')
+        type = v.get('type', int)
+        v['value'] = type(value)
+
+
+def resolve_kwargs(global_kwargs, kwargs):
+    # not quite perfect hack, so we can be verbose here
+    verbose = kwargs.get('verbose', 0) or global_kwargs['verbose']
+
+    system_kwargs = {}
+    other_kwargs = {}
+
+    for k in kwargs:
+        if k in global_kwargs:
+            gkw = global_kwargs[k]
+            if gkw.get('strong'):
+                if verbose:
+                    print('environment variable overrides passed in value for', k, file=sys.stderr)
+                system_kwargs[k] = gkw['value']
+            else:
+                system_kwargs[k] = kwargs[k]
+        else:
+            other_kwargs[k] = kwargs[k]
+    for k in global_kwargs:
+        if k not in system_kwargs:
+            system_kwargs[k] = global_kwargs[k]['value']
+
+    return system_kwargs, other_kwargs
 
 
 def make_subdir_name(count, prefix='ps'):
