@@ -3,15 +3,19 @@ import os
 import tempfile
 import pytest
 import sys
+from io import StringIO
 
 import paramsurvey
 import paramsurvey.stats
 from paramsurvey.examples import sleep_worker, burn_worker
 
 
+pslogger_fd = StringIO()
+
+
 @pytest.fixture(scope="session")
 def paramsurvey_init(request):
-    paramsurvey.init()
+    paramsurvey.init(pslogger_fd=pslogger_fd)
 
     def finalize():
         # needed to get pytest multiprocessing coverage
@@ -20,6 +24,14 @@ def paramsurvey_init(request):
     request.session.addfinalizer(finalize)
 
 
+def readouterr_and_dump(capsys):
+    captured = capsys.readouterr()
+    # this dumping is useful when there are failures to debug
+    sys.stdout.write('Dumped stdout:\n' + captured.out)
+    sys.stderr.write('Dumped stderr:\n' + captured.err)
+    return captured
+
+
 def test_basics(paramsurvey_init):
     ncores = max(4, paramsurvey.current_core_count())
 
@@ -108,32 +120,28 @@ def test_args(capsys, paramsurvey_init):
     assert test_user_kwargs.get('out_func_called')
     assert len(results) == 2
 
-    captured = capsys.readouterr()
-    sys.stdout.write(captured.out)
-    sys.stderr.write(captured.err)
+    captured = readouterr_and_dump(capsys)
     assert len(captured.err.splitlines()) >= len(psets)
 
     # because of progress_dt being 0., we should have at least len(psets) progress lines
     has_name = [line for line in captured.err.splitlines() if 'progress' in line and name in line]
     assert len(has_name) >= len(psets)
 
-    # same as previous but verbose=2 instead of progress_dt=0
-    name = 'test_args verbose=2'
+    # same as previous but verbose=3 instead of progress_dt=0
+    name = 'test_args verbose=3'
     results = paramsurvey.map(do_test_args, psets,
                               out_func=out_func, user_kwargs=test_user_kwargs,
                               chdir=chdir, outfile=outfile, out_subdirs=10,
-                              verbose=2, name=name)
+                              verbose=3, name=name)
 
     assert out_func_called
     assert test_user_kwargs.get('out_func_called')
     assert len(results) == 2
 
-    captured = capsys.readouterr()
-    sys.stdout.write(captured.out)
-    sys.stderr.write(captured.err)
+    captured = readouterr_and_dump(capsys)
     assert len(captured.err.splitlines()) >= len(psets)
 
-    # because of progress_dt being 0., we should have at least len(psets) progress lines
+    # because of verbose=3, we should have at least len(psets) progress lines
     has_name = [line for line in captured.err.splitlines() if 'progress' in line and name in line]
     assert len(has_name) >= len(psets)
 
@@ -150,18 +158,18 @@ def do_raise(pset, system_kwargs, user_kwargs, raw_stats):
 def test_worker_exception(capsys, paramsurvey_init):
     psets = [{'raises': False}, {'raises': False}, {'raises': False}, {'raises': True}, {'raises': False}, {'raises': False}, {'raises': False}]
 
+    pslogger_fd.seek(0)
     results = paramsurvey.map(do_raise, psets, name='test_worker_exception')
     assert len(results) == 6
     assert len(results.missing) == 1
     assert '_exception' in results.missing[0]
+    assert '_traceback' in results.missing[0]
     assert results.progress.total == 7
     assert results.progress.finished == 6
     assert results.progress.failures == 1
     assert results.progress.exceptions == 1
 
-    captured = capsys.readouterr()
-    sys.stdout.write(captured.out)
-    sys.stderr.write(captured.err)
+    captured = readouterr_and_dump(capsys)
 
     # ray redirects stderr to stdout, while multiprocessing prints it in the worker
     # TODO: add stderr/out capture everywhere and use it here
@@ -170,6 +178,11 @@ def test_worker_exception(capsys, paramsurvey_init):
     # the standard progress function prints this
     assert 'failures: 1' in captured.out or 'failures: 1' in captured.err
 
+    log = pslogger_fd.getvalue()
+    assert 'Traceback' in log, 'python traceback seen'
+    assert 'ValueError' in log, 'python exception seen'
+    assert 'pset:' in log, 'our pset logline seen'
+
 
 def do_nothing(pset, system_kwargs, user_kwargs, raw_stats):
     return {'foo': True}
@@ -187,6 +200,8 @@ def test_wrapper_exception(capsys, paramsurvey_init):
     assert results.progress.failures == 2
 
     # ray and multiprocessing behave differently for wrapper 'exception'
+    # multiprocessing has exception and traceback, both in results.missing and pslogger
+    # our ray wrapper doesn't even try to catch it? does it print 'Surprised'?
     #assert '_exception' in results.missing[0]
     #assert results.progress.exceptions == 1
 
@@ -194,9 +209,7 @@ def test_wrapper_exception(capsys, paramsurvey_init):
 
     # XXX ray prints traceback in the worker, multiprocessing and ray local_mode prints in the parent
 
-    captured = capsys.readouterr()
-    sys.stdout.write(captured.out)
-    sys.stderr.write(captured.err)
+    captured = readouterr_and_dump(capsys)
     assert 'failures: 2' in captured.out or 'failures: 2' in captured.err
 
 
@@ -224,5 +237,12 @@ def test_toplevel(paramsurvey_init):
     assert paramsurvey.backend() == os.environ.get('PARAMSURVEY_BACKEND', paramsurvey.default_backend)
 
 
+def test_kwargs(paramsurvey_init):
+    duration = 0.1
+    psets = [{'duration': duration}] * 10
+    results = paramsurvey.map(sleep_worker, psets, limit=3, name='sleep with limit')
+    assert len(results) == 3
+
+
 def test_overlarge_pset():
     pass
